#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <queue>
#include "squarecmdline.h"
#include <random>
#include <vector>
#include <fstream>
#include <iomanip>

using namespace std;
char debuglevel;

class coor{
public:
  double x;
  double y;
  coor(double a,double b){
    x=a;
    y=b;
  }
  coor(){
  }
};

ostream& operator<<(ostream& os, const coor& point)
{
  os << "("<< point.x << "," << point.y << ")";
  return os;
}

bool equal(double a, double b){
  return (a==b || abs(a-b)<abs(min(a,b))*numeric_limits<double>::epsilon());
}

bool equal(coor p, coor q){
  return (equal(p.x,q.x) && equal(p.y,q.y));
}


double edist(coor p, coor q){
  double edistsq=(p.x-q.x)*(p.x-q.x)+(p.y-q.y)*(p.y-q.y);
  return sqrt(edistsq);
}

class vertex{
public:
  coor pos;
  double cost;
  //  bool seen;
  int parent; //index of parent in explored[]
  bool updatecost(double cost0, coor endx, coor endy, coor target){
    double mycost=cost0+edist(endx,endy)+edist(target,endy)-edist(target,endx);
    if (mycost<cost){
      cost = mycost;
      return true;
    }
    return false;
  }
  void setcost(double cost0, coor endx, coor endy, coor target){
    double mycost=cost0+edist(endx,endy)+edist(target,endy)-edist(target,endx);
    cost = mycost;
  }
  vertex(){
    //    seen=false;
  }
  vertex(coor p, double c, int prev){
    pos=p;
    cost=c;
    //    seen=true;
    parent=prev;
  }
  /*  vertex& operator=(const vertex& source){
    pos=source.pos;
    seen=source.seen;
    cost=source.cost;
    parent=source.parent;
    return *this;
    }*/
};

bool operator < (vertex v, vertex u){
  return v.cost<u.cost;
}

bool operator > (vertex v, vertex u){
  return v.cost>u.cost;
}

bool operator >= (vertex v, vertex u){
  return v.cost>=u.cost;
}

bool operator <= (vertex v, vertex u){
  return v.cost<=u.cost;
}




class vxheap{
  bool hasarray;
public:
  vector<vertex> array;
  int size(){
    return array.size();
  }
private:
  bool siftdown(int l)
  {
    int size=array.size();
    bool moved=false;
    int j;
    vertex x;
    j = 2*l+1;
    x = array[l];
    while (j < size)
      {
	if ( j <size - 1)
	  if ( array[j] > array[j+1])
	    ++j;
	if (x <= array[j])
	  break;
	array[l] = array[j];
	l = j;
	j = 2*l + 1; // sift
	moved=true;
      }
    array[l] = x;
    return moved;
  }

  bool siftup(int l)
  {
    //    int i,j;
    //    vertex x;
    //    i = l;
    //    j = 2*i+1;
    bool moved=false;
    int j;
    vertex x = array[l];
    while (l >0)
      {
	j=(l-1)/2;
	if (x >= array[j])
	  break;
	else{
	  array[l] = array[j];
	  l=j;
	}
	moved=true;
      }
    array[l] = x;
    return moved;
  }
public:
  void fix(int l){
    if (!siftup(l))
      siftdown(l);
  }
  void insert(vertex vx){
    array.push_back(vx);
    siftup(array.size()-1);
  }

  vertex pop(){
    vertex vx=array[0];
    array[0]=array[array.size()-1];
    array.pop_back();
    siftdown(0);
    return vx;
  }
  
  vxheap(){
  }
};

void topstricks(string fileprefix, vector<vertex>&explored, coor START, coor FINISH){
  ofstream LATEXfile;
  char filename[200] ;
  sprintf(filename,"%s.tex",fileprefix.c_str());
  LATEXfile.open (filename);

  LATEXfile << fixed << showpoint;
  LATEXfile << setprecision(6);

  LATEXfile << "\%Compile with:\n";
  LATEXfile << "\%latex geodesic.tex;  dvips -E -f -Ppdf geodesic.dvi > geodesic.eps; epstopdf geodesic.eps\n";
  LATEXfile << "\%Total path length is sqrt(2)+"<<explored.back().cost<<endl;
  LATEXfile << "\\documentclass{article}\n";
  LATEXfile << "\\usepackage{pstricks,pst-eps,xcolor,graphicx,kpfonts}\n";
  LATEXfile << "\\begin{document}\n";
  LATEXfile << "\\begin{TeXtoEPS}\n";
  LATEXfile << "\\psset{unit=4cm, linewidth=.5pt,dimen=outer,dotsize=5pt}\n";
  LATEXfile << "\\begin{pspicture}(0,0)(1,1)"<< endl;
  LATEXfile << "\\psframe[linewidth=.1pt,linestyle=dashed](0,0)(1,1)"<< endl;

  int travel=explored.size()-1;
  int pathdist=0;
  LATEXfile << "\\psline" << explored[travel].pos;
  while (!equal(explored[travel].pos,START)){
    travel=explored[travel].parent;
    LATEXfile << explored[travel].pos;
    pathdist++;
  }
  LATEXfile << endl;
  LATEXfile << "\\end{pspicture}\n";
  LATEXfile << "\\end{TeXtoEPS}\n";
  LATEXfile << "\\end{document}\n";
  LATEXfile.close();
  cout << "Path length is "<<pathdist<<endl;
  
}

int main(int argc, char* argv[])
{
  gengetopt_args_info lineArgs;
  if (cmdline_parser(argc, argv, &lineArgs)) {
    cmdline_parser_print_help();
    return 1;
  }


  coor START=coor(lineArgs.px_arg,lineArgs.py_arg);
  coor FINISH=coor(lineArgs.qx_arg,lineArgs.qy_arg);


  //  cerr << "size of vertex is "<<sizeof(vertex)<<endl;

  int64_t n=pow(2,lineArgs.vxs_arg);
  int64_t nleft=n;

  double p=lineArgs.avgdegree_arg/n;

  //  default_random_engine gen;
  mt19937_64 gen(314159265358979);

  cout << "n is "<<n<<endl;
  
  //  cout <<p<<endl;
  //  vertex * explored=new vertex[MAXSET];
  //  explored[0]=vertex(coor(0,0),0,0);
  

  n--; //take out vx at 0,0.
  vxheap frontier;
  vector<vertex> explored;
  //  explored.reserve(MAXSET);
  explored.push_back(vertex(START,0,0));

  uniform_real_distribution<> undist(0,1);
  int neighbors;
  {
    binomial_distribution<int64_t> bindist(n,p);
    neighbors=bindist(gen);
  }
  for (int i=0; i<neighbors; i++){
    coor pos(undist(gen),undist(gen));
    vertex vx;
    vx.pos=pos;
    vx.setcost(0,START,pos,FINISH);
    vx.parent = 0;
    frontier.insert(vx);
    nleft--;
  }


  int count=0;
  while (frontier.size()>0){
    if (0 && !(count%1000000)){
      //      cerr << "n used is .. " << n-nleft<<endl;
      cerr << "heap size is "<<frontier.size()<<endl;
    }
    //    cerr << count << " " << frontier.size() <<endl;
    count++; //safety for now
    explored.push_back(frontier.pop());
    if (equal(explored.back().pos,FINISH)){
      cout << "DONE! shortest path has length error "<< explored.back().cost<<endl;
      topstricks(lineArgs.filename_arg,explored, START, FINISH);
      return 0;
    }
    int newneighbors;
    {
      binomial_distribution<int64_t> bindist(nleft,p);
      newneighbors=bindist(gen);
    }
    for (int i=0; i<newneighbors; i++){
      vertex newvx;
      if (undist(gen)<1./nleft) //hit the target!
	newvx.pos=FINISH;
      else
	newvx.pos=coor(undist(gen),undist(gen));
      newvx.setcost(explored.back().cost,explored.back().pos,newvx.pos,FINISH);
      newvx.parent=(explored.size()-1);
      if ((!lineArgs.max_cost_given) || newvx.cost<=lineArgs.max_cost_arg)
	frontier.insert(newvx);
      nleft--;
    }
    int numfneighbors;
    {
      binomial_distribution<int64_t> bindist(frontier.size(),p);
      numfneighbors=bindist(gen);
    }

    vector<int> fneighbors(numfneighbors);
    std::uniform_int_distribution<int> intdist(0,frontier.size()-1);
    for (int i=0; i<numfneighbors; i++){
    retry:
      int vindex=intdist(gen);
      for (int j=0; j<i; j++){
	if (vindex==fneighbors[j]){
	  //	  cerr << "retrying"<<endl;
	  goto retry;
	}
      }
      fneighbors[i]=vindex;
    }
	
    for (int i=0; i<numfneighbors; i++){
      int vindex=fneighbors[i];
      if (frontier.array[vindex].updatecost(explored.back().cost,explored.back().pos,frontier.array[vindex].pos,FINISH)){
	frontier.array[vindex].parent=explored.size()-1;
	frontier.fix(vindex);
      }
    }
  }
  if (lineArgs.max_cost_given)
    cout << "No path with extra cost < " << lineArgs.max_cost_arg<< " exists!"<<endl;
  else
    cout << "NOT CONNETED! "<< nleft<<endl;


  //  priority_queue frontier
}
